from typing import AsyncGenerator
from litellm import ModelResponse, completion
from typing import Callable, Optional

from just_agents.memory import Memory
from just_agents.streaming.abstract_streaming import AbstractStreaming, FunctionParser
from just_agents.streaming.protocols.abstract_protocol import AbstractStreamingProtocol
from just_agents.streaming.protocols.openai_streaming import OpenaiStreamingProtocol
import json
from qwen_agent.llm import get_chat_model
import litellm


class Qwen2AsyncSession(AbstractStreaming):

    def __init__(self, llm_session, output_streaming: AbstractStreamingProtocol = OpenaiStreamingProtocol()):
        super().__init__(llm_session)
        self.output_streaming = output_streaming


    def _process_function(self, name: str, arguments: str, available_tools: dict[str, Callable]):
        function_args = json.loads(arguments)
        function_to_call = available_tools[name]
        print(function_to_call, function_args)
        try:
            function_response = function_to_call(**function_args)
        except Exception as e:
            function_response = str(e)
        message = {
            'role': 'function',
            'name': name,
            'content': function_response,
        }  # TODO need to track arguments , arguments=function_args
        return message

    async def resp_async_generator(self
                                   ) -> AsyncGenerator[str, None]:
        from just_agents.llm_session import LLMSession
        llm_session: LLMSession = self.session
        llm = get_chat_model(llm_session.llm_options)
        functions = None
        if llm_session.available_tools:
            functions = []
            for fun_name in llm_session.available_tools:
                functions.append(litellm.utils.function_to_dict(llm_session.available_tools[fun_name]))

        proceed = True
        while proceed:
            proceed = False
            responces = llm.chat(messages=llm_session.memory.messages, functions=functions, stream=True,
                                 extra_generate_cfg=dict(parallel_function_calls=True))
            prev_len = 0
            messages = []
            for i, msg in enumerate(responces):
                messages = msg
                content = msg[0]["content"]
                if len(content) > 0:
                    delta = content[prev_len:]
                    prev_len = len(content)
                    yield self.output_streaming.get_chunk(i, delta, llm_session.llm_options)

            fncall_msgs = [rsp for rsp in messages if rsp.get('function_call', None)]
            llm_session.memory.add_messages(messages)
            # function_call = messages[-1].get('function_call', None)
            if fncall_msgs and len(fncall_msgs) > 0 and llm_session.available_tools:
                proceed = True
                for msg in fncall_msgs:
                    function_call = msg['function_call']
                    llm_session.memory.add_message(self._process_function(function_call["name"], function_call['arguments'], llm_session.available_tools))
        yield self.output_streaming.done()


from just_agents.streaming.protocols.abstract_protocol import AbstractStreamingProtocol
import json
import time


class OpenaiStreamingProtocol(AbstractStreamingProtocol):

    def get_chunk(self, index: int, delta: str, options: dict):
        chunk = {
            "id": index,
            "object": "chat.completion.chunk",
            "created": time.time(),
            "model": options["model"],
            "choices": [{"delta": {"content": delta}}],
        }
        return f"data: {json.dumps(chunk)}\n\n"


    def done(self):
        return "data: [DONE]\n\n"


class AbstractStreamingProtocol:

    def get_chunk(self, index:int, delta:str, options:dict):
        raise NotImplementedError()
        return ""

    def done(self):
        raise NotImplementedError()
        return ""


from dataclasses import dataclass, field
from typing import Callable, Optional
from litellm.types.utils import Function

OnMessageCallable = Callable[[dict], None]
OnFunctionCallable = Callable[[Function], None]

@dataclass
class Memory:

    on_message: list[OnMessageCallable] = field(default_factory=list)
    messages: list[dict] = field(default_factory=list)

    def add_on_message(self, handler: OnMessageCallable):
        self.on_message.append(handler)

    def add_on_tool_result(self, handler: OnMessageCallable):
        self.add_on_message(lambda m: handler(m) if m['role'] == "tool" else None)

    def add_on_tool_call(self, fun: OnFunctionCallable):
        """
        Adds handler only to function calls to track what exactly was called
        :param fun:
        :return:
        """
        def tool_handler(message: dict) -> None:
            if hasattr(message, 'tool_calls') and message.tool_calls is not None:
                for call in message.tool_calls:
                    #if call.function is Function:
                    fun(call.function)
        self.add_on_message(tool_handler)


    def remove_on_message(self, handler: OnMessageCallable):
        self.on_message = [m for m in self.on_message if m == handler]

    def add_system_message(self, prompt: str):
        return self.add_message({"role":"system", "content":prompt})

    def add_user_message(self, prompt: str):
        return self.add_message({"role":"user", "content":prompt})

    def add_message(self, message: dict):
        """
        adds message to the memory
        :param message:
        :param run_callbacks:
        :return:
        """
        self.messages.append(message)
        for handler in self.on_message:
            handler(message)

    @property
    def last_message(self) -> Optional[dict]:
        return self.messages[-1] if len(self.messages) > 0 else None


    def add_messages(self, messages: list[dict]):
        for message in messages:
            self.messages.append(message)
            for handler in self.on_message:
                handler(message)

    def clear_messages(self):
        self.messages.clear()



#from just_agents.chat_agent import ChatAgent
import json
import pprint

import pytest
from dotenv import load_dotenv

import just_agents.llm_options
from just_agents.llm_session import LLMSession
import asyncio
from tests.mock import *


def get_current_weather(location: str):
    """Gets the current weather in a given location"""
    if "tokyo" in location.lower():
        return json.dumps({"location": "Tokyo", "temperature": "10", "unit": "celsius"})
    elif "san francisco" in location.lower():
        return json.dumps({"location": "San Francisco", "temperature": "72", "unit": "fahrenheit"})
    elif "paris" in location.lower():
        return json.dumps({"location": "Paris", "temperature": "22", "unit": "celsius"})
    else:
        return json.dumps({"location": location, "temperature": "unknown"})

def test_sync_function_calling():
    load_dotenv(override=True)
    session: LLMSession = LLMSession(
        llm_options=just_agents.llm_options.LLAMA3_2,
        tools=[get_current_weather]
    )
    result = session.query("What's the weather like in San Francisco, Tokyo, and Paris?")
    assert "72" in result
    assert "22" in result
    assert "10" in result

async def process_stream(async_generator):
    async for item in async_generator:
        pass

def test_stream_function_calling():
    load_dotenv(override=True)
    session: LLMSession = LLMSession(
        llm_options=just_agents.llm_options.LLAMA3_2,
        tools=[get_current_weather]
    )
    stream = session.stream("What's the weather like in San Francisco, Tokyo, and Paris?")
    loop = asyncio.get_event_loop()
    loop.run_until_complete(process_stream(stream))
    result = session.memory.last_message["content"]
    assert "72" in result
    assert "22" in result
    assert "10" in result


def test_stream_genetics_function_calling():
    load_dotenv(override=True)
    session: LLMSession = LLMSession(
        llm_options=just_agents.llm_options.LLAMA3_2,
        tools=[hybrid_search, rsid_lookup, gene_lookup, pathway_lookup, disease_lookup, sequencing_info, clinical_trails_full_trial]
    )
    stream = session.stream("What is the influence of different alleles in rs10937739?")
    loop = asyncio.get_event_loop()
    loop.run_until_complete(process_stream(stream))
    result = session.memory.last_message["content"]
    assert "pro-longevity" in result

@pytest.mark.skip(reason="so far qwen inference we are using has issues with json function calling")
def test_async_gwen2_function_calling():
    load_dotenv(override=True)
    session: LLMSession = LLMSession(
        llm_options=just_agents.llm_options.OPEN_ROUTER_Qwen_2_72B_Instruct,
        tools=[get_current_weather]
    )
    result = session.query_all_messages("What's the weather like in San Francisco, Tokyo, and Paris?")
    assert "72" in result
    assert "22" in result
    assert "10" in result



from just_agents_sandbox.micromamba_session import MicromambaSession
from llm_sandbox.docker import ConsoleOutput

def run_bash_command(command: str):
    """
    command: str # command to run in bash, for example install software inside micromamba environment
    """
    with  MicromambaSession(image="ghcr.io/longevity-genie/just-agents/biosandbox:main", lang="python", keep_template=True, verbose=True) as session:
        result: ConsoleOutput = session.execute_command(command=command)
        return result
    

def run_python_code(code: str):
    """
    code: str # python code to run in micromamba environment
    """
    with  MicromambaSession(image="ghcr.io/longevity-genie/just-agents/biosandbox:main", lang="python", keep_template=True, verbose=True) as session:
        result: ConsoleOutput = session.run(code)
        return result
    
def copy_files_from_runtime(src: str, dest: str):
    """
    src: str # path to file in runtime
    dest: str # path to file in host
    """
    with  MicromambaSession(image="ghcr.io/longevity-genie/just-agents/biosandbox:main", lang="python", keep_template=True, verbose=True) as session:
        result: ConsoleOutput = session.copy_from_runtime(src=src, dest=dest)
        return result


def copy_files_to_runtime(src: str, dest: str):
    """
    src: str # path to file in host
    dest: str # path to file in runtime
    """
    with  MicromambaSession(image="ghcr.io/longevity-genie/just-agents/biosandbox:main", lang="python", keep_template=True, verbose=True) as session:
        result: ConsoleOutput = session.copy_to_runtime(src=src, dest=dest)
        return result


from pathlib import Path
from typing import AsyncGenerator, Any
from just_agents.utils import resolve_agent_schema

def build_agent(agent_schema: str | Path | dict):
    from just_agents.cot_agent import ChainOfThoughtAgent
    from just_agents.llm_session import LLMSession
    agent_schema = resolve_agent_schema(agent_schema)
    class_name = agent_schema.get("class", None)
    if class_name is None or class_name == "LLMSession":
        return LLMSession(agent_schema=agent_schema)
    elif class_name == "ChainOfThoughtAgent":
        return ChainOfThoughtAgent(agent_schema=agent_schema)

class IAgent:
    def stream(self, input: str | dict | list[dict]) -> AsyncGenerator[Any, None]:
        raise NotImplementedError("You need to impelement stream() first!")

    def query(self, input: str | dict | list[dict]) -> str:
        raise NotImplementedError("You need to impelement query() first!")


from typing import Optional
import docker
from llm_sandbox.session import SandboxDockerSession
from llm_sandbox.docker import ConsoleOutput
from llm_sandbox.const import SupportedLanguage

class MicromambaSession(SandboxDockerSession):
    def __init__(self, client: Optional[docker.DockerClient] = None,
        image: Optional[str] = None,
        dockerfile: Optional[str] = None,
        lang: str = SupportedLanguage.PYTHON,
        keep_template: bool = False,
        verbose: bool = False, environment: str = "base"):
        super().__init__(client=client,
                         image=image,
                         dockerfile=dockerfile,
                         lang=lang,
                         keep_template=keep_template,
                         verbose=verbose)
        self.environment = environment


    def execute_command(
        self, command: Optional[str], workdir: Optional[str] = None
    ) -> ConsoleOutput:
        if not command:
            raise ValueError("Command cannot be empty")

        if not self.container:
            raise RuntimeError(
                "Session is not open. Please call open() method before executing commands."
            )
        command = f"micromamba run -n {self.environment} {command}"

        if self.verbose:
            print(f"Executing command: {command}")

        if workdir:
            exit_code, exec_log = self.container.exec_run(
                command, stream=True, tty=True, workdir=workdir
            )
        else:
            exit_code, exec_log = self.container.exec_run(
                command, stream=True, tty=True
            )

        output = ""
        if self.verbose:
            print("Output:", end=" ")

        for chunk in exec_log:
            chunk_str = chunk.decode("utf-8")
            output += chunk_str
            if self.verbose:
                print(chunk_str, end="")

        return ConsoleOutput(output)



from typing import Any, Dict, Optional

from just_agents.llm_session import LLMSession
from just_agents.utils import load_config


class ChatAgent(LLMSession):
    """
    Default agent class, it assumes that each agent has a role, goal, and background story.
    So far this class is in a refactoring process, use LLMSession instead.
    """
    def __init__(self, role: str, goal: Optional[str] = None, task: Optional[str] = None,
                 backstory: Optional[str] = None, config: Dict[str, Any] = None,
                 agent_prompt: Optional[str] = None, **kwargs):
        super().__init__(**kwargs)
        self.role = role
        self.goal = goal
        self.task = task
        self.backstory = backstory
        self.config = config or load_config("agent_prompts.yaml")
        self.agent_prompt = agent_prompt

        agent_template: str = str(self.config["agent_prompt"])
        self.agent_prompt = self._render_template(agent_template)
        if self.agent_prompt is not None:
            self.instruct(self.agent_prompt)

    def _render_template(self, template: str):
        """
        renders agent template
        :param template:
        :return:
        """
        return template.format(role=self.role,
                               goal=self.goal,
                               backstory=self.backstory,
                               task=self.task)



from typing import AsyncGenerator

from litellm import ModelResponse, completion
from typing import Callable, Optional

from just_agents.memory import Memory
from just_agents.streaming.abstract_streaming import AbstractStreaming, FunctionParser
from just_agents.streaming.protocols.abstract_protocol import AbstractStreamingProtocol
from just_agents.streaming.protocols.openai_streaming import OpenaiStreamingProtocol


class AsyncSession(AbstractStreaming):

    def __init__(self, llm_session, output_streaming: AbstractStreamingProtocol = OpenaiStreamingProtocol()):
        super().__init__(llm_session)
        self.output_streaming = output_streaming


    async def resp_async_generator(self) -> AsyncGenerator[str, None]:
        from just_agents.llm_session import LLMSession
        llm_session: LLMSession = self.session
        proceed = True
        while proceed:
            proceed = False
            response: ModelResponse = llm_session._rotate_completion(stream=True)
            parser: Optional[FunctionParser] = None
            tool_messages: list[dict] = []
            parsers: list[FunctionParser] = []
            deltas: list[str] = []
            for i, part in enumerate(response):
                delta: str = part["choices"][0]["delta"].get("content")  # type: ignore
                if delta:
                    deltas.append(delta)
                    yield self.output_streaming.get_chunk(i, delta, llm_session.llm_options)

                tool_calls = part["choices"][0]["delta"].get("tool_calls")
                if tool_calls and (llm_session.available_tools is not None):
                    if not parser:
                        parser = FunctionParser(id = tool_calls[0].id)
                    if parser.parsed(tool_calls[0].function.name, tool_calls[0].function.arguments):
                        tool_messages.append(self._process_function(parser, llm_session.available_tools))
                        parsers.append(parser)
                        parser = None #maybe Optional?

            if len(tool_messages) > 0:
                proceed = True
                llm_session.memory.add_message(self._get_tool_call_message(parsers))
                for message in tool_messages:
                    llm_session.memory.add_message(message)

            if len(deltas) > 0:
                llm_session.memory.add_message({"role":"assistant", "content":"".join(deltas)})

        yield self.output_streaming.done()


import asyncio
from dotenv import load_dotenv
from just_agents.cot_agent import ChainOfThoughtAgent
import pytest

import just_agents.llm_options

def count_letters(character:str, word:str) -> str:
    """ Returns the number of character occurrences in the word. """
    count:int = 0
    for char in word:
        if char == character:
            count += 1
    print("Function: ", character, " occurres in ", word, " ", count, " times.")
    return str(count)


def test_function_query():
    load_dotenv(override = True)

    opt = just_agents.llm_options.OPENAI_GPT4oMINI.copy()
    agent: ChainOfThoughtAgent = ChainOfThoughtAgent(opt, tools=[count_letters])
    result, thoughts = agent.query("Count the number of occurrences of the letter ’L’ in the word - ’LOLLAPALOOZA’.")
    print("Thoughts: ", thoughts)
    print("Results: ", result)
    assert "4" in result

async def process_stream(async_generator):
    for item in async_generator:
        print(item)


def test_stream_function_query():
    load_dotenv(override = True)

    opt = just_agents.llm_options.OPENAI_GPT4oMINI.copy()
    agent: ChainOfThoughtAgent = ChainOfThoughtAgent(opt, tools=[count_letters])
    stream = agent.stream("Count the number of occurrences of the letter ’L’ in the word - ’LOLLAPALOOZA’.")
    loop = asyncio.get_event_loop()
    loop.run_until_complete(process_stream(stream))
    result = agent.last_message()
    print(result)
    assert "4" in str(result["content"])


from typing import List

def hybrid_search(query: str) -> str:
    """Mock function for hybrid search."""
    print(f"hybrid_search({query})")
    return f"Mocked hybrid search result for query: {query}"

def rsid_lookup(rsid: str) -> str:
    """Retrieves information about rsid. It could be the effects of longevity or its influence on diseases.
     rsid is identifier for single nucleotide polymorphism, and looks like rs123 where instead of 123 could be any integer."""
    print(f"rsid_lookup({rsid})")
    return f"{rsid} has pro-longevity influence"

def gene_lookup(gene: str) -> str:
    """Mock function for gene lookup."""
    print(f"gene_lookup({gene})")
    return f"Mocked information for gene: {gene}"

def pathway_lookup(pathway: str) -> str:
    """Mock function for pathway lookup."""
    print(f"pathway_lookup({pathway})")
    return f"Mocked information for pathway: {pathway}"

def disease_lookup(disease: str) -> str:
    """Mock function for disease lookup."""
    print(f"disease_lookup({disease})")
    return f"Mocked information for disease: {disease}"

def sequencing_info() -> str:
    """Mock function for sequencing info."""
    print("sequencing_info()")
    return "Mocked DNA sequencing information"

def _process_sql(sql_query: str) -> List[dict]:
    """Mock function for processing SQL query."""
    print(f"_process_sql({sql_query})")
    return [{"mocked_key": "mocked_value"}]

def clinical_trails_full_trial(trial_id: str) -> dict:
    """Mock function for full clinical trial details."""
    print(f"clinical_trails_full_trial({trial_id})")
    return {"trial_id": trial_id, "mocked_detail": "This is a mocked clinical trial detail"}




import random

class RotateKeys():
    keys:list[str]

    def __init__(self, file_path:str):
        with open(file_path) as f:
            text = f.read().strip()
            self.keys = text.split("\n")

    def __call__(self, *args, **kwargs):
        return random.choice(self.keys)

    def remove(self, key:str):
        self.keys.remove(key)

    def len(self):
        return len(self.keys)


import pprint
from typing import Optional, Union
import requests
from typing import List, Dict, Any

from semanticscholar import SemanticScholar
from semanticscholar.PaginatedResults import PaginatedResults
from semanticscholar.Paper import Paper

def get_semantic_paper(query: str,
                       year: str = None,
                       publication_types: list = None,
                       open_access_pdf: bool = None,
                       venue: list = None,
                       fields_of_study: list = None,
                       fields: list = None,
                       publication_date_or_year: str = None,
                       min_citation_count: int = None,
                       limit: int = 100,
                       bulk: bool = False,
                       sort: str = None):
    """
    Search for academia papers by keyword.

        :param str query: plain-text search query string.
        :param str year: (optional) restrict results to the given range of \
               publication year.
        :param list publication_types: (optional) restrict results to the given \
               publication type list.
        :param bool open_access_pdf: (optional) restrict results to papers \
               with public PDFs.
        :param list venue: (optional) restrict results to the given venue list.
        :param list fields_of_study: (optional) restrict results to given \
               field-of-study list, using the s2FieldsOfStudy paper field.
        :param list fields: (optional) list of the fields to be returned.
        :param str publication_date_or_year: (optional) restrict results to \
               the given range of publication date in the format \
               <start_date>:<end_date>, where dates are in the format \
               YYYY-MM-DD, YYYY-MM, or YYYY.
        :param int min_citation_count: (optional) restrict results to papers \
               with at least the given number of citations.
        :param int limit: (optional) maximum number of results to return \
               (must be <= 100).
        :param bool bulk: (optional) bulk retrieval of basic paper data \
               without search relevance (ignores the limit parameter if True \
               and returns up to 1,000 results in each page).
        :param str sort: (optional) sorts results (only if bulk=True) using \
               <field>:<order> format, where "field" is either paperId, \
               publicationDate, or citationCount, and "order" is asc \
               (ascending) or desc (descending).
        :returns: query results
    """
    sch = SemanticScholar()
    results: PaginatedResults = sch.search_paper(query,
                               year=year,
                               publication_types=publication_types,
                               open_access_pdf=open_access_pdf,
                               venue=venue,
                               fields_of_study=fields_of_study,
                               fields=fields,
                               publication_date_or_year=publication_date_or_year,
                               min_citation_count=min_citation_count,
                               limit=limit,
                               bulk=bulk,
                               sort=sort
                               )
    return results


def hybrid_search(text: str,
                  collections = ["aging_papers_paragraphs_bge_base_en_v1.5", "aging_papers_paragraphs_specter2"],
                  limit: int = 10,
                  db: str = "https://localhost:9200",
                  verbose: bool = False,
                  host: str = "https://api.longevity-genie.info", string: bool = True) -> Union[str, list[str]]:
    """
    Searching in academic literature. Note: do not change default parameters for db and collections unless there are explicit users or systems instructions

    Parameters:
    - text (str): The text to search for.
    - collections (List[str]): A list of collection names to search within
    - limit (int): The maximum number of results to return.
    - db (str): The database URL to query.
    - verbose (bool): Whether to include verbose output. Default is False.
    - host (str): The host URL for the search endpoint. Default is "https://agingkills.eu".
    - string (str): if result should be concatenated to str and not list

    Returns:
    - List[str]: response from the hybrid search API.
    """

    # Endpoint for the hybrid search
    endpoint = f"{host}/hybrid_search"

    # Payload to send in the POST request
    payload = {
        "text": text,
        "collections": collections,
        "limit": limit,
        "db": db,
        "verbose": verbose
    }

    # Perform the POST request
    response = requests.post(endpoint, json=payload)

    # Raise an exception if the request was unsuccessful
    response.raise_for_status()

    # Return the JSON response
    results = response.json()
    return ".".join(results).replace("\\n", "\n") if string else results


def literature_search(query: str, limit: int = 10):
    """
    Search in the academic literature
    :param query:
    :param limit: limit number of results
    :return: extracts from scientific papers where SOURCE field is the id of the paper
    """
    return hybrid_search(text=query, limit=limit)



from just_agents.interfaces.IAgent import IAgent
from just_agents.llm_session import LLMSession
import json
from just_agents.streaming.protocols.openai_streaming import OpenaiStreamingProtocol
from just_agents.streaming.protocols.abstract_protocol import AbstractStreamingProtocol
from pathlib import Path, PurePath
import yaml
from just_agents.utils import _resolve_agent_schema, resolve_llm_options, resolve_system_prompt, resolve_tools

# schema parameters:
LLM_SESSION = "llm_session"
THOUGHT_MAX_TOKES = "thought_max_tokes"
CONTENT_NAME = "content"
MAX_STEPS = "max_steps"
NEXT_ACTION = "next_action"
ACTION_FINAL = "action_final"
FINAL_MAX_TOKENS = "final_max_tokens"
FINAL_PROMPT = "final_prompt"


class ChainOfThoughtAgent(IAgent):

    def __init__(self, llm_options: dict = None, agent_schema: str | Path | dict | None = None,
                 tools: list = None, output_streaming:AbstractStreamingProtocol = OpenaiStreamingProtocol()):
        self.agent_schema: dict = _resolve_agent_schema(agent_schema, "cot_agent_prompt.yaml")
        if tools is None:
            tools = resolve_tools(self.agent_schema)
        self.session: LLMSession = LLMSession(llm_options=resolve_llm_options(self.agent_schema, llm_options),
                                              system_prompt=resolve_system_prompt(self.agent_schema),
                                              agent_schema=self.agent_schema.get(LLM_SESSION, None), tools=tools)

        self.output_streaming: AbstractStreamingProtocol = output_streaming


    def stream(self, input: str | dict | list):
        thought_max_tokes = self.agent_schema.get(THOUGHT_MAX_TOKES, 300)
        self.session.update_options("max_tokens", thought_max_tokes)
        self.session.update_options("response_format", {"type": "json_object"})
        step_data = json.loads(self.session.query(input))
        CONTENT = self.agent_schema.get(CONTENT_NAME, "content")
        content = step_data[CONTENT] + "\n"
        yield self.output_streaming.get_chunk(0, content, self.session.llm_options)
        max_steps = self.agent_schema.get(MAX_STEPS, 25)
        for step_count in range(1, max_steps):
            step_data = json.loads(self.session.proceed())
            content = step_data[CONTENT] + "\n"
            yield self.output_streaming.get_chunk(step_count, content, self.session.llm_options)
            if step_data[self.agent_schema.get(NEXT_ACTION, "next_action")] == self.agent_schema.get(ACTION_FINAL, "final_answer"):
                break

        final_max_tokens = self.agent_schema.get(FINAL_MAX_TOKENS, 1200)
        self.session.update_options("max_tokens", final_max_tokens)
        final_prompt = self.agent_schema.get(FINAL_PROMPT, "Please provide the final answer based solely on your reasoning above.")
        final_data = json.loads(self.session.query(final_prompt))
        yield self.output_streaming.get_chunk(step_count + 1, final_data[CONTENT], self.session.llm_options)
        yield self.output_streaming.done()


    def query(self, input:  str | dict | list):
        thought_max_tokes = self.agent_schema.get(THOUGHT_MAX_TOKES, 300)
        self.session.update_options("max_tokens", thought_max_tokes)
        self.session.update_options("response_format", {"type": "json_object"})
        step_data = json.loads(self.session.query(input))
        CONTENT = self.agent_schema.get(CONTENT_NAME, "content")
        content = step_data[CONTENT] + "\n"
        thoughts:str = content
        max_steps = self.agent_schema.get(MAX_STEPS, 25)
        for step_count in range(1, max_steps):
            step_data = json.loads(self.session.proceed())
            content = step_data[CONTENT] + "\n"
            thoughts += content
            if step_data[self.agent_schema.get(NEXT_ACTION, "next_action")] == self.agent_schema.get(ACTION_FINAL, "final_answer"):
                break

        final_max_tokens = self.agent_schema.get(FINAL_MAX_TOKENS, 1200)
        self.session.update_options("max_tokens", final_max_tokens)
        final_prompt = self.agent_schema.get(FINAL_PROMPT, "Please provide the final answer based solely on your reasoning above.")
        final_data = json.loads(self.session.query(final_prompt))
        return final_data[CONTENT], thoughts

    def last_message(self):
        return self.session.memory.last_message


from dataclasses import dataclass, field, replace
from dataclasses import asdict
from typing import Any, Dict, Optional


from dataclasses import dataclass, asdict

OPENAI_GPT4o: Dict = {
    "model": "gpt-4o",
    "temperature": 0.0
}

OPENAI_GPT4oMINI: Dict = {
    "model": "gpt-4o-mini",
    "temperature": 0.0
}


LLAMA3: Dict = {
    "model": "groq/llama3-groq-70b-8192-tool-use-preview",
    "api_base": "https://api.groq.com/openai/v1",
    "temperature": 0.0,
    "tools": []
}

LLAMA3_1: Dict = {
    "model": "groq/llama-3.1-70b-versatile",
    "api_base": "https://api.groq.com/openai/v1",
    "temperature": 0.0,
    "tools": []
}

LLAMA3_2_VISION: Dict = {
    "model": "groq/llama-3.2-11b-vision-preview",
    "api_base": "https://api.groq.com/openai/v1",
    "temperature": 0.0,
    "tools": []
}

LLAMA3_2: Dict = {
    "model": "groq/llama-3.2-90b-text-preview",
    "api_base": "https://api.groq.com/openai/v1",
    "temperature": 0.0,
    "tools": []
}

OPEN_ROUTER_Qwen_2_72B_Instruct = {
    "model": "openrouter/qwen/qwen-2-72b-instruct",
    "temperature": 0.0,
    "tools": []
}

OPEN_ROUTER_Qwen_2_72B_Instruct_Vision = {
    "model": "openrouter/qwen/qwen-2-vl-72b-instruct",
    "temperature": 0.0,
    "tools": []
}

OPEN_ROUTER_LLAMA_3_8B_FREE = {
    "model": "openrouter/meta-llama/llama-3-8b-instruct:free",
    "temperature": 0.0,
    "tools": []
}

OPEN_ROUTER_GEMINI_1_5_FLASH_EXP_FREE = {
    "model": "openrouter/google/gemini-flash-1.5-exp",
    "temperature": 0.0,
    "tools": []
}

MISTRAL_8x22B = {
    "model": "mistral/open-mixtral-8x22b",
    "temperature": 0.0
}


import json
import time
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Dict, Callable, AsyncGenerator, Optional
from just_agents.streaming.protocols.abstract_protocol import AbstractStreamingProtocol

from just_agents.memory import Memory


@dataclass
class FunctionParser:
    id: str = ""
    name: str = ""
    arguments: str = ""

    def parsed(self, name: str, arguments: str):
        if name:
            self.name += name
        if arguments:
            self.arguments += arguments
        if len(self.name) > 0 and len(self.arguments) > 0 and self.arguments.strip().endswith("}"):
            return True
        return False


class AbstractStreaming(ABC):
    """
    Class that is required to implement the streaming logic
    """
    output_streaming: AbstractStreamingProtocol

    def __init__(self, llm_session):
        self.session = llm_session

    @abstractmethod
    async def resp_async_generator(self) -> AsyncGenerator:
        """
        Async generator that fills memory with streaming data
        :param memory:
        :param options:
        :param available_tools:
        :return:
        """
        pass

    def _process_function(self, parser: FunctionParser, available_tools: Dict[str, Callable]):
        function_args = json.loads(parser.arguments)
        function_to_call = available_tools[parser.name]
        try:
            function_response = function_to_call(**function_args)
        except Exception as e:
            function_response = str(e)
        message = {"role":"tool", "content":function_response, "name":parser.name,
                         "tool_call_id":parser.id}  # TODO need to track arguments , arguments=function_args
        return message

    def _get_tool_call_message(self, parsers: list[FunctionParser]) -> dict:
        tool_calls = []
        for parser in parsers:
            tool_calls.append({"type":"function",
                "id": parser.id, "function": {"name": parser.name, "arguments": parser.arguments}})
        return {"role":"assistant", "content":None, "tool_calls":tool_calls}


from setuptools import setup, find_packages
import codecs
import os

here = os.path.abspath(os.path.dirname(__file__))

with codecs.open(os.path.join(here, "README.md"), encoding="utf-8") as fh:
    long_description = "\n" + fh.read()

VERSION = '0.2.1'
DESCRIPTION = 'Just Agents'
LONG_DESCRIPTION = 'LLM Agents that are implemented without unnecessary complexity'

# Setting up
setup(
    name="just-agents",
    version=VERSION,
    author="Alex Karmazin, Anton Kulaga",
    author_email="antonkulaga@gmail.com",
    description=DESCRIPTION,
    long_description_content_type="text/markdown",
    long_description=long_description,
    packages=find_packages(),
    install_requires=["litellm>=1.50.1", "numpydoc", "requests", "typer>=0.12.5"],
    extras_require={
        'tools': [
            # subproject with useful tools for agents
            'semanticscholar>=0.8.4'
        ]
    },
    keywords=['python', 'llm', 'science', 'review', 'agents', 'AI'],
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        "Operating System :: Unix",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
    ],
    package_data={
        'just_agents': ['config/*.yaml'],
        'just_agents_sandbox': ['config/*.yaml'],
        
    },
    include_package_data=True,
)



import copy
import json
from pathlib import Path
from typing import Any, AsyncGenerator

import litellm
from litellm import ModelResponse, completion
from litellm.utils import Choices

from just_agents.interfaces.IAgent import IAgent
from just_agents.llm_options import LLAMA3
from just_agents.memory import Memory
from dataclasses import dataclass, field
from typing import Callable, Optional
from just_agents.streaming.abstract_streaming import AbstractStreaming
from just_agents.streaming.openai_streaming import AsyncSession
# from just_agents.utils import rotate_completion
from just_agents.utils import _resolve_agent_schema, resolve_llm_options, resolve_system_prompt, resolve_tools
from just_agents.rotate_keys import RotateKeys

OnCompletion = Callable[[ModelResponse], None]

# schema parameters:
KEY_LIST_PATH = "key_list_path"
STREAMING_METHOD = "streaming_method"
BACKUP_OPTIONS = "backup_options"
COMPLETION_REMOVE_KEY_ON_ERROR = "completion_remove_key_on_error"
COMPLETION_MAX_TRIES = "completion_max_tries"

#streaming methods:
OPENAI = "openai"
QWEN2 = "qwen2"
CHAIN_OF_THOUGHT = "chain_of_thought"


class LLMSession(IAgent):

    def __init__(self, llm_options: dict[str, Any] = None,
                 system_prompt:str = None,
                 agent_schema: str | Path | dict | None = None,
                 tools: Optional[list[Callable]] = None):
        self.on_response = []
        self.available_tools: Optional[dict[str, Callable]] = {}
        self.memory: Memory = Memory()
        self.streaming: AbstractStreaming = None
        self.key_getter: RotateKeys = None
        self.on_response: list[OnCompletion] = []

        self.agent_schema = _resolve_agent_schema(agent_schema, "llm_session_schema.yaml")
        self.llm_options: dict[str, Any] = resolve_llm_options(self.agent_schema, llm_options)
        if self.agent_schema.get(KEY_LIST_PATH, None) is not None:
            self.key_getter = RotateKeys(self.agent_schema[KEY_LIST_PATH])
        self.tools: list[Callable] = tools
        if self.tools is None:
            self.tools = resolve_tools(self.agent_schema)

        if self.llm_options is not None:
            self.llm_options = copy.deepcopy(self.llm_options) #just a satefy requirement to avoid shared dictionaries
            if (self.key_getter is not None) and (self.llm_options.get("api_key", None) is not None):
                print("Warning api_key will be rewriten by key_getter. Both are present in llm_options.")

        if system_prompt is None:
            system_prompt = resolve_system_prompt(self.agent_schema)
        if system_prompt is not None:
            self.instruct(system_prompt)

        streaming_method = self.agent_schema.get(STREAMING_METHOD, None)
        if streaming_method is None or streaming_method == OPENAI:
            self.streaming = AsyncSession(self)
        elif streaming_method.lower() == QWEN2:
            from just_agents.streaming.qwen2_streaming import Qwen2AsyncSession
            self.streaming = Qwen2AsyncSession(self)
        elif streaming_method.lower() == CHAIN_OF_THOUGHT:
            from just_agents.streaming.chain_of_thought import ChainOfThought
            self.streaming = ChainOfThought(self)
        else:
            raise ValueError("just_streaming_method is incorrect. "
                             "It should be one of this ['qwen2', 'chain_of_thought']")

        if self.tools is not None:
            self._prepare_tools(self.tools)


    def add_on_response_listener(self, listener:OnCompletion):
        self.on_response.append(listener)


    def remove_on_response_listener(self, listener:OnCompletion):
        if listener in self.on_response:
            self.on_response.remove(listener)


    def _rotate_completion(self, stream: bool) -> ModelResponse:
        opt = self.llm_options.copy()
        max_tries = self.agent_schema.get(COMPLETION_MAX_TRIES, 2)
        if self.key_getter is not None:
            if max_tries < 1:
                max_tries = self.key_getter.len()
            else:
                if self.agent_schema.get(COMPLETION_REMOVE_KEY_ON_ERROR, True):
                    max_tries = min(max_tries, self.key_getter.len())
            last_exception = None
            for _ in range(max_tries):
                opt["api_key"] = self.key_getter()
                try:
                    response = completion(messages=self.memory.messages, stream=stream, **opt)
                    return response
                except Exception as e:
                    last_exception = e
                    if self.agent_schema.get(COMPLETION_REMOVE_KEY_ON_ERROR, True):
                        self.key_getter.remove(opt["api_key"])

            backup_opt: dict = self.agent_schema.get(BACKUP_OPTIONS, None)
            if backup_opt:
                return completion(messages=self.memory.messages, stream=stream, **backup_opt)
            if last_exception:
                raise last_exception
            else:
                raise Exception(
                    f"Run out of tries to execute completion. Check your keys! Keys {self.key_getter.len()} left.")
        else:
            return completion(messages=self.memory.messages, stream=stream, **opt)


    def _process_response(self, response: ModelResponse):
        """
        Running handlers to process model responses
        :param response:
        :return:
        """
        for handler in self.on_response:
            handler(response)


    @staticmethod
    def message_from_response(response: ModelResponse) -> dict:
        choice: Choices = response.choices[0]
        message: dict = choice.message.to_dict(exclude_none=True, exclude_unset=True) #converting to dictionary and deleting nonw
        if "function_call" in message and message["function_call"] is None:
            del message["function_call"]
        return message


    def instruct(self, prompt: str):
        """
        Ads system instructions
        :param prompt:
        :return:
        """
        system_instruction =  {"content":prompt, "role":"system"}
        self.memory.add_message(system_instruction)
        return system_instruction


    def update_options(self, key:str, value:Any):
        self.llm_options[key] = value


    def _add_to_memory(self, input: str | dict | list[dict]):
        if isinstance(input, list):
            self.memory.add_messages(input)
        else:
            if isinstance(input, str):
                input = {"role": "user", "content": input}
            self.memory.add_message(input)


    def query(self, input: str | dict | list[dict]) -> str:
        """
        Query large language model
        :param input:
        :return:
        """
        self._add_to_memory(input)
        return self.proceed()


    def stream(self, input: str | dict | list[dict]) -> AsyncGenerator[Any, None]: # -> ContentStream:
        """
        streaming method
        :input prompt, message, list of messages:
        :return:
        """
        self._add_to_memory(input)
        return self.proceed_stream()


    def proceed_stream(self) -> AsyncGenerator[Any, None]: # -> ContentStream:
        return self.streaming.resp_async_generator()


    def proceed(self) -> str:
        response: ModelResponse = self._rotate_completion(stream=False)
        self._process_response(response)
        response = self._process_function_calls(response)
        answer = self.message_from_response(response)
        self.memory.add_message(answer)
        result: str = self.memory.last_message["content"] if "content" in self.memory.last_message else str(self.memory.last_message)
        return result


    def _process_function_calls(self, response: ModelResponse) -> Optional[ModelResponse]:
        """
        processes function calls in the response
        :param response_message:
        :return:
        """
        proceed = True
        while proceed:
            proceed = False
            response_message = response.choices[0].message
            tool_calls = response_message.get("tool_calls")

            if tool_calls and (self.tools is not None):
                proceed = True
                message = self.message_from_response(response)
                self.memory.add_message(message)
                for tool_call in tool_calls:
                    function_name = tool_call.function.name
                    function_to_call = self.available_tools[function_name]
                    function_args = json.loads(tool_call.function.arguments)
                    try:
                        function_response = str(function_to_call(**function_args))
                    except Exception as e:
                        function_response = str(e)
                    result = {"role":"tool", "content":function_response, "name":function_name, "tool_call_id":tool_call.id}
                    self.memory.add_message(result)
                response = self._rotate_completion(stream=False)
                self._process_response(response)
        return response

    def _prepare_tools(self, functions: list[Any]):
        """
        Prepares functions as tools that LLM can call.
        Note, the functions should have comments explaining LLM how to use them
        :param functions:
        :return:
        """
        tools = []
        self.available_tools = {}
        for fun in functions:
            function_description = litellm.utils.function_to_dict(fun)
            self.available_tools[function_description["name"]] = fun
            tools.append({"type": "function", "function": function_description})
        self.llm_options["tools"] = tools
        self.llm_options["tool_choice"] = "auto"


from typing import AsyncGenerator

from litellm import ModelResponse, completion
from typing import Callable, Optional

from just_agents.memory import Memory
from just_agents.streaming.abstract_streaming import AbstractStreaming, FunctionParser
from just_agents.streaming.protocols.openai_streaming import OpenaiStreamingProtocol
from just_agents.streaming.protocols.abstract_protocol import AbstractStreamingProtocol
import time
import json

class ChainOfThought(AbstractStreaming):

    def __init__(self, llm_session, output_streaming: AbstractStreamingProtocol = OpenaiStreamingProtocol()):
        super().__init__(llm_session)
        self.output_streaming = output_streaming


    async def resp_async_generator(self) -> AsyncGenerator[str, None]:
        print("This method depricated use cot_agent instead.")
        from just_agents.llm_session import LLMSession
        llm_session: LLMSession = self.session
        max_steps = 25

        opt = llm_session.llm_options.copy()
        opt["max_tokens"] = 300
        opt["response_format"] = {"type": "json_object"}
        for step_count in range(1, max_steps):
            response = llm_session._rotate_completion(stream=False)
            step_data = json.loads(response.choices[0].message.content)
            llm_session.memory.add_message({"role": "assistant", "content": json.dumps(step_data)})
            print(step_count, " ", step_data)
            content = step_data['content'] + "\n"
            yield self.output_streaming.get_chunk(step_count, content, opt)
            if step_data[
                'next_action'] == 'final_answer':  # Maximum of 25 steps to prevent infinite thinking time. Can be adjusted.
                break

        llm_session.memory.add_message({"role": "user",
                            "content": "Please provide the final answer based solely on your reasoning above."})

        opt["max_tokens"] = 1200
        response = llm_session._rotate_completion(stream=False)
        final_data = json.loads(response.choices[0].message.content)
        # yield steps, total_thinking_time
        print("Final: ", final_data)
        yield self.output_streaming.get_chunk(step_count + 1, final_data['content'], opt)
        yield self.output_streaming.done()


import os
import random

import yaml
from pathlib import Path
from typing import Optional, Dict, Any
import importlib.resources as resources
from dotenv import load_dotenv
import importlib
from typing import Callable
from litellm import Message, ModelResponse, completion
import copy

#
# class RotateKeys():
#     keys:list[str]
#
#     def __init__(self, file_path:str):
#         with open(file_path) as f:
#             text = f.read().strip()
#             self.keys = text.split("\n")
#
#     def __call__(self, *args, **kwargs):
#         return random.choice(self.keys)
#
#     def remove(self, key:str):
#         self.keys.remove(key)
#
#     def len(self):
#         return len(self.keys)

def resolve_agent_schema(agent_schema: str | Path | dict):
    """
    Resolve the agent schema from a string, path (to yaml file) or dict.
    """
    if isinstance(agent_schema, str):
        agent_schema = Path(agent_schema)
    if isinstance(agent_schema, Path):
        if not agent_schema.exists():
            raise ValueError(
                f"In constructor agent_schema path is not exists: ({str(agent_schema)})!")
        with open(agent_schema) as f:
            agent_schema = yaml.safe_load(f)
    if not isinstance(agent_schema, dict):
        raise ValueError(
            f"In constructor agent_schema parameter should be string, Path or dict!")

    return agent_schema

def _resolve_agent_schema(agent_schema: str | Path | dict | None, default_file_name: str):
    if agent_schema is None:
        agent_schema = Path(Path(__file__).parent, "config", default_file_name)

    return resolve_agent_schema(agent_schema)


def resolve_llm_options(agent_schema: dict, llm_options: dict):
    if llm_options is None:
        llm_options = agent_schema.get("options", None)
    if llm_options is None:
        raise ValueError(
            "llm_options should not be None. You should pass it through llm_options or agent_schema['options'].")

    return llm_options


def resolve_system_prompt(agent_schema: dict):
    system_prompt = agent_schema.get("system_prompt", None)
    system_prompt_path = agent_schema.get("system_prompt_path", None)
    if (system_prompt is not None) and (system_prompt_path is not None):
        raise ValueError("You should use only one of system_prompt or system_prompt_path not both together.")
    if system_prompt_path is not None:
        path = Path(system_prompt_path)
        if path.exists():
            system_prompt = path.read_text(encoding="utf8")
    return system_prompt


def resolve_tools(agent_schema: dict) -> list[Callable]:
    function_list:list[Callable] = []
    tools = agent_schema.get('tools', None)
    if tools is None:
        return None
    for entry in tools:
        package_name: str = entry['package']
        function_name: str = entry['function']
        try:
            # Dynamically import the package
            package = importlib.import_module(package_name)
            # Get the function from the package
            func = getattr(package, function_name)
            function_list.append(func)
        except (ImportError, AttributeError) as e:
            print(f"Error importing {function_name} from {package_name}: {e}")

    return function_list


def rotate_env_keys() -> str:
    load_dotenv()
    keys = []
    index = 1

    while True:
        # Determine the environment variable name
        key_name = 'KEY' if index == 0 else f'KEY_{index}'
        key_value = os.getenv(key_name)

        # Break the loop if the key is not found
        if key_value is None:
            break

        # Add the found key to the list
        keys.append(key_value)
        index += 1

    # Raise an error if no keys are found
    if not keys:
        raise ValueError("No keys found in environment variables")

    # Randomly choose one of the available keys
    return random.choice(keys)

def load_config(resource: str, package: str = "just_agents.config") -> Dict[str, Any]:
    """
    :rtype: yaml config
    """
    if Path(resource).exists():
        return yaml.safe_load(Path(resource).open("r"))
    in_config = Path("config") / resource
    if in_config.exists():
        return yaml.safe_load( (Path("config") / resource).open("r") )
    else:
        # Load from package resources
        with resources.open_text(package, 'agent_prompts.yaml') as file:
            return yaml.safe_load(file)